#include <vector>
#include <cstdio>
#include <string>
#include <list>
#include <algorithm>

struct LengthEvent
{
    int length;        // -length = triplet speed
    int note;          // note*12+oct,  0=pause

    bool operator== (const LengthEvent& e) const { return length==e.length && note==e.note; }
};
struct CommandEvent
{
    char type;         // ADHSV
    unsigned param:16; // byte or two
    
    bool operator== (const CommandEvent& e) const { return type==e.type && param==e.param; }
};
struct Event
{
    bool cmd;
    union
    {
        LengthEvent l;
        CommandEvent c;
    };
    
    bool operator==(const Event& e) const
    {
        if(cmd != e.cmd) return false;
        if(cmd)
            return c == e.c;
        return l == e.l;
    }
};
struct Track
{
    std::vector<Event> events;
    unsigned loopbegin;
    
    Track(): loopbegin(0) { }
    
    void SetLoopBegin() { loopbegin=events.size(); }
    
    void SequenceCommand(const CommandEvent& c) const
    {
        switch(c.type)
        {
            case 'A':
                printf(" .byt MUSCMD_SetSpeed,        $%02X\n", c.param); break;
            case 'C':
                printf(" .byt MUSCMD_SetVibratoCtrl,  $%02X\n", c.param); break;
            case 'D':
                printf(" .byt MUSCMD_SetDutyCycle ,   $%02X\n", c.param); break;
            case 'E':
                printf(" .byt MUSCMD_SetNoteDelay,    $%02X\n", c.param); break;
            case 'H':
                printf(" .byt MUSCMD_SetVibratoIndex, $%02X\n", c.param); break;
            case 'S':
                printf(" .byt MUSCMD_SetVolumeCurve,  $%02X,$%02X\n", c.param>>8, c.param&0xFF); break;
            case 'V':
                printf(" .byt MUSCMD_SetVolumeAndEnv, $%02X\n", c.param); break;
            default:
                fprintf(stderr, "Unknown command: `%c%X'\n", c.type, c.param);
        }
    }
    void SequenceNote(long length, int note) const
    {
        long orig_length = length;
        
        if(length < 0)
        {
            long len_trip = -length;
            if(!(len_trip%3))
            {
                length = len_trip*2/3;
            }
        }
        
        if(length != 0)
        {
            //printf(" ;length %ld\n", length);
            const char* pref = "    ";
            long l = length;
            if(l < 0) { l=-l; pref="$30,"; }
        
        eatmore:
            if(l == 96) { printf(" .byt %s6,$%02X\n", pref,0xE0+note); l-=96; goto eatmore; }
            if(l == 48) { printf(" .byt %s6,$%02X\n", pref,0xC0+note); l-=48; goto eatmore; }
            if(l == 24) { printf(" .byt %s6,$%02X\n", pref,0xA0+note); l-=24; goto eatmore; }
            if(l == 12) { printf(" .byt %s6,$%02X\n", pref,0x80+note); l-=12; goto eatmore; }
            if(l ==  6) { printf(" .byt %s6,$%02X\n", pref,0x60+note); l-= 6; goto eatmore; }
            if(l ==  3) { printf(" .byt %s6,$%02X\n", pref,0x40+note); l-= 3; goto eatmore; }
            while(l >= 64) { printf(" .byt %s  $%02X\n", pref,0xE0+note); l-=64; goto eatmore; }
            while(l >= 32) { printf(" .byt %s  $%02X\n", pref,0xC0+note); l-=32; goto eatmore; }
            while(l >= 16) { printf(" .byt %s  $%02X\n", pref,0xA0+note); l-=16; goto eatmore; }
            while(l >=  8) { printf(" .byt %s  $%02X\n", pref,0x80+note); l-= 8; goto eatmore; }
            while(l >=  4) { printf(" .byt %s  $%02X\n", pref,0x60+note); l-= 4; goto eatmore; }
            while(l >=  2) { printf(" .byt %s  $%02X\n", pref,0x40+note); l-= 2; goto eatmore; }
            
            if(length > 0) length=l; else length=-l;
        }
        if(length != 0)
            fprintf(stderr, "Unsupported length remainder: %ld from original length %ld\n", length, orig_length);
    }
    void SequencePause(long len_full, long len_trip) const
    {
        //std::fprintf(stderr, "SequencePause(%ld,%ld)\n", len_full, len_trip);
        
        if(!(len_trip % 3))
        {
            len_full += len_trip*2/3;
            len_trip = 0;
        }
        else
        {
            len_trip += len_full*3/2;
            len_full = 0;
        }
        
        SequenceNote(len_full,  0);
        SequenceNote(-len_trip, 0);
    }
    void CalcNewBase(unsigned& valid, long& value, unsigned begin, unsigned end) const
    {
        valid = 0;
        value = 0;
        
        long min = 999;
        long max = 0;
        
        for(unsigned pos=begin; pos<end; ++pos)
        {
            const Event& e = events[pos];
            if(e.cmd)     { ok: ++valid; continue; }
            if(!e.l.note) goto ok;
            
            if(e.l.note > max)
            {
                long newmax = e.l.note;
                if(newmax - min > 0x1F) break;
                max = newmax;
            }
            if(e.l.note < min)
            {
                long newmin = e.l.note;
                if(max - newmin > 0x1F) break;
                min = newmin;
            }
            goto ok;
        }
        if(max == 0) return;
        
        value = min;
        printf(" .byt MUSCMD_SetNoteBaseOffset, %ld*12%+ld\n", min/12, min%12-1);
    }
    void SequenceRegion(unsigned begin, unsigned end) const
    {
        unsigned base_valid = 0;
        long     base_value = 0;
        
        long pauselen_full = 0;
        long pauselen_trip = 0;
        #define FlushPause() \
                if(pauselen_full || pauselen_trip) \
                { \
                    SequencePause(pauselen_full, pauselen_trip); \
                    pauselen_full = pauselen_trip = 0; \
                }
        for(unsigned pos=begin; pos<end; ++pos)
        {
            if(!base_valid)
            {
                CalcNewBase(base_valid, base_value, pos, end);
            }
            --base_valid;

            const Event& e = events[pos];
            if(e.cmd)
            {
                FlushPause();
                SequenceCommand(e.c);
            }
            else if(e.l.note)
            {
                FlushPause();
                SequenceNote(e.l.length, e.l.note - base_value + 1);
            }
            else
            {
                if(e.l.length < 0)
                    pauselen_trip += -e.l.length;
                else
                    pauselen_full += e.l.length;
            }
        }
        FlushPause();
        #undef FlushPause
    }
    
    struct CompressionInfo
    {
        unsigned begin;
        unsigned lines;
        unsigned count;
        unsigned end;  // = begin + lines*count
    };
    void FindBiggestCompression(unsigned begin, unsigned end, CompressionInfo& info) const
    {
        unsigned best_begin   = begin;
        unsigned best_length  = end-begin;
        unsigned best_count   = 1;
        
        unsigned best_score = 0;
        
        for(unsigned pos=begin; pos<end; ++pos)
        {
            unsigned max_length = (end-pos) / 2;
            for(unsigned test_length = 1; test_length < max_length; ++test_length)
            {
                unsigned count = 1;
                while(count < 126)
                {
                    unsigned matchpos = pos + test_length * count;
                    unsigned matchend = matchpos + test_length;
                    if(matchend > end) break;
                    
                    if(!std::equal(&events[pos], &events[pos+test_length],
                                   &events[matchpos])) break;
                      
                    ++count;
                }
                unsigned saving = test_length * (count-1);
                if(saving > 3 && saving > best_score)
                {
                    best_score  = saving;
                    best_begin  = pos;
                    best_length = test_length;
                    best_count  = count;
                }
            }
        }
        
        info.begin = best_begin;
        info.lines = best_length;
        info.count = best_count;
        info.end   = info.begin + info.lines*info.count;
    }
    void CompileRegion(unsigned begin, unsigned end, bool is_main_loop = false) const
    {
        CompressionInfo info;
        FindBiggestCompression(begin, end, info);
        
        bool abuse_main_loop = is_main_loop;
        
        if(info.begin > begin)
        {
            CompileRegion(begin, info.begin);
            begin = info.begin;
            abuse_main_loop = false;
        }
        
        if(info.count > 1 && !abuse_main_loop)
        {
            std::printf("-\n");
        }
        SequenceRegion(info.begin, info.begin + info.lines);
        if(info.count > 1)
        {
            std::printf(" .byt MUSCMD_Loop, %u-1 : .word %s\n",
                info.count, abuse_main_loop ? "--" : "-");
        }
        if(info.end < end)
        {
            CompileRegion(info.end, end);
        }
    }
    void Compile() const
    {
        bool has_loop = loopbegin > 0;
        unsigned useloopbegin = loopbegin;
        
        unsigned mainloopend = events.size();
        
        if(has_loop)
        {
            CompressionInfo info;
            FindBiggestCompression(loopbegin, mainloopend, info);
            
            if(info.begin == loopbegin && info.end == mainloopend)
            {
                /* When it's the main loop, it doesn't matter how many times it exactly
                 * is looped. Take that particular region only and subdivide.
                 * Example:
                 *   { abc abc abc abc } *inf
                 * Instead of saying:
                 *   { { abc } *4  } *inf
                 * We can say simply:
                 *   { abc } *inf
                 */
                mainloopend = loopbegin + info.lines;
            }

            FindBiggestCompression(0, mainloopend, info);
            
            /* Find whether the main loop can begin earlier */
            if(info.count > 1 && info.end == mainloopend && info.begin < loopbegin)
            {
                useloopbegin = info.begin;
                mainloopend = info.begin + info.lines;
            }
        }
        
        if(useloopbegin > 0)
        {
            CompileRegion(0, useloopbegin);
        }
        if(has_loop) std::printf("--\n");
        
        CompileRegion(useloopbegin, mainloopend, has_loop);
        
        if(has_loop)
            std::printf(" .byt MUSCMD_Loop, 0 : .word --\n");
        else
            std::printf(" .byt MUSCMD_EndOfData\n");
    }
};
struct Song
{
    Track channels[4];
    std::string name;
    
    void Compile()
    {
        std::printf(
            "MUS_%s:\n"
            ".(\n"
            " .byt 15\n"
            " .word chn0\n"
            " .word chn1\n"
            " .word chn2\n"
            " .word chn3\n"
            " .word vibtab\n", name.c_str());
        for(unsigned c=0; c<4; ++c)
        {
            std::printf("chn%u\n", c);
            channels[c].Compile();
        }
        std::printf(
            "vibtab\n"
            ".)\n");
    }
};

struct BookMark
{
    long pos;
    explicit BookMark(long p) : pos(p) { }
    void Use() { std::fseek(stdin, pos, SEEK_SET); }
private:
    BookMark();
};
BookMark GetBookMark() { return BookMark(std::ftell(stdin)); }
#define EOFMARK "EOF"
#define NEWLINE    "\n"
#define WHITE      " "
#define LOOPBEGIN  "{"
#define LOOPEND    "}"
#define COLON      ":"
#define NEXTOCT    ">"
#define PREVOCT    "<"
static int Peekc() { int c = std::getchar(); std::ungetc(c, stdin); return c; }
static const std::string Scan()
{
    int c = std::getchar();
    if(c == EOF) return EOFMARK;
    if(c == '\n') return NEWLINE;
    if(c == ' ' || c == '\t' || c == '\r') return WHITE;
    if(c == '{') return LOOPBEGIN;
    if(c == '}') return LOOPEND;
    if(c == '<') return PREVOCT;
    if(c == '>') return NEXTOCT;
    
    if(c >= 'a' && c <= 'z')
    {
        std::string note;
        note += (char) c;
        if(Peekc() == '#') { std::getchar(); note += '#'; }
        return note; 
    }
    if((c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9'))
    {
        std::string tok;
        std::ungetc(c, stdin);
        while((c >= 'A' && c <= 'Z') || (c >= '0' && c <= '9'))
        {
            c = std::getchar();
            tok += (char)c;
            c = Peekc();
        }
        return tok;
    }
    if(c == ':') return COLON;
    std::fprintf(stderr, "Parse error '%c'\n", c);
    throw "Parse error";
}
static long ScanNum()
{
    BookMark pos = GetBookMark();
    std::string tok = Scan();
    if(tok[0] < '0' || tok[0] > '9') { pos.Use(); return -1; }
    return std::atoi(tok.c_str());
}
static void SkipBlank()
{
redo:
    while(Peekc() == ' ') std::getchar();
    
    BookMark p = GetBookMark();
    int c1 = std::getchar();
    int c2 = std::getchar();
    if(c1 == '/' && c2 == '*')
    {
        c2 = std::getchar();
        for(;;)
        {
            c1 = c2;
            c2 = std::getchar();
            if(c2 == EOF) break;
            if(c1 == '*' && c2 == '/') break;
        }
        goto redo;
    }
    p.Use();
}
static void SkipCommentline()
{
retry:
    SkipBlank();
    switch(Peekc())
    {
        case ';': for(;;) { int c = std::getchar(); if(c == EOF || c == '\n') goto retry; }
    }
}
static long ScanLength(bool optional=true)
{
    long len = ScanNum();
    if(len < 0) { if(optional) return 0; throw "Len missing"; }
    if(Peekc() == 't') { len = -len; std::getchar(); }
    return len;
}

struct Loop
{
    BookMark pos;
    long done;
    Loop() : pos(GetBookMark()), done(0) { }
};

static Song ReadInput()
{
    Song result;
    
    unsigned curchn = 0;
    
    std::list<Loop> loops;
    
    long len[4] = {1,1,1,1};
    long oct[4] = {2,2,2,2};
    
    SkipCommentline();
    for(;;)
    {
        long c = ScanNum();
        SkipBlank();
        if(c >= 0)
        {
            if(Scan() != COLON) throw "Parse error after number";
            curchn = c;
            if(curchn >= 4)
            {
                std::fprintf(stderr, "Channel %u is not available\n", curchn);
                throw "Parse error";
            }
            continue;
        }
        std::string tok = Scan();
        if(tok == EOFMARK) break;
        if(tok == LOOPBEGIN)
        {
            loops.push_back(Loop());
            continue;
        }
        if(tok == LOOPEND)
        {
            long count = ScanNum();
            if(count < 0) throw "Loop number missing";
            Loop& l = *loops.begin();
            if(++l.done < count)
            {
                l.pos.Use();
                continue;
            }
            loops.erase(loops.begin());
            continue;
        }
        if(tok == NEWLINE) { SkipCommentline(); continue; }
        if(tok == WHITE) continue;
        if(tok == NEXTOCT) { ++oct[curchn]; continue; }
        if(tok == PREVOCT) { --oct[curchn]; continue; }
        if(tok == "l")
        {
            len[curchn] = ScanLength(false);
            continue;
        }
        if(tok == "o")
        {
            oct[curchn] = ScanLength(false);
            continue;
        }
        if(tok == "LOOPBEGIN")
        {
            for(unsigned c=0; c<4; ++c)
                result.channels[c].SetLoopBegin();
            continue;
        }
        
        #define OPTLEN(target) do{ long l = ScanLength(true); target = l ? l : len[curchn]; }while(0)
        #define STORE(ev) result.channels[curchn].events.push_back(ev);
        
        if(tok == "p")
        {
            long pauselen;
            OPTLEN(pauselen);
            while(pauselen != 0)
            {
                Event e;
                e.cmd = false;
                e.l.note = 0;
                e.l.length = pauselen;
                pauselen=0;
               // e.l.length = pauselen < 0 ? -1 : 1;
               // if(pauselen < 0) ++pauselen; else --pauselen;
                STORE(e);
            }
            continue;
        }
        if(tok[0] >= 'A' && tok[0] <= 'Z')
        {
            Event e;
            e.cmd = true;
            e.c.type  = tok[0];
            char* endptr;
            e.c.param = strtol(tok.c_str()+1, &endptr, 16);
            if(*endptr)
            {
                std::fprintf(stderr, "Unrecognized command: `%s'\n", tok.c_str());
            }
            STORE(e);
            continue;
        }

        Event e;
        e.cmd = false;
        if(tok == "c") e.l.note = 0;
        else if(tok == "c#") e.l.note = 1;
        else if(tok == "d") e.l.note = 2;
        else if(tok == "d#") e.l.note = 3;
        else if(tok == "e") e.l.note = 4;
        else if(tok == "f") e.l.note = 5;
        else if(tok == "f#") e.l.note = 6;
        else if(tok == "g") e.l.note = 7;
        else if(tok == "g#") e.l.note = 8;
        else if(tok == "a") e.l.note = 9;
        else if(tok == "a#") e.l.note = 10;
        else if(tok == "b") e.l.note = 11;
        else
        {
            std::fprintf(stderr, "Unrecognized note: `%s'\n", tok.c_str());
            throw "parse error";
        }
        e.l.note = oct[curchn]*12 + e.l.note;
        OPTLEN(e.l.length);
        STORE(e);
    }
    return result;
}

int main(int argc, const char* const* argv)
{
    Song song = ReadInput();

    if(argc == 2) song.name = argv[1];
    
    song.Compile();
}
